@@ -1,7 +1,7 @@
 /*
  * android.c - Setup IPC for shadowsocks-android
  *
- * Copyright (C) 2013 - 2018, Max Lv <max.c.lv@gmail.com>
+ * Copyright (C) 2013 - 2019, Max Lv <max.c.lv@gmail.com>
  *
  * This file is part of the shadowsocks-libev.
  *
@@ -92,9 +92,12 @@ protect_socket(int fd)
     return ret;
 }
 
+extern char *stat_path;
+
 int
 send_traffic_stat(uint64_t tx, uint64_t rx)
 {
+    if (!stat_path) return 0;
     int sock;
     struct sockaddr_un addr;
 
@@ -112,7 +115,7 @@ send_traffic_stat(uint64_t tx, uint64_t rx)
 
     memset(&addr, 0, sizeof(addr));
     addr.sun_family = AF_UNIX;
-    strncpy(addr.sun_path, "stat_path", sizeof(addr.sun_path) - 1);
+    strncpy(addr.sun_path, stat_path, sizeof(addr.sun_path) - 1);
 
     if (connect(sock, (struct sockaddr *)&addr, sizeof(addr)) == -1) {
         LOGE("[android] connect() failed for stat_path: %s (socket fd = %d)\n",
@@ -129,14 +132,6 @@ send_traffic_stat(uint64_t tx, uint64_t rx)
         return -1;
     }
 
-    char ret = 0;
-
-    if (recv(sock, &ret, 1, 0) == -1) {
-        ERROR("[android] recv");
-        close(sock);
-        return -1;
-    }
-
     close(sock);
-    return ret;
+    return 0;
 }
